import os
import signal
import time
from asyncio import AbstractEventLoop, get_event_loop
from typing import Any, Optional

import typer

from asyncapi import (
    AsyncApi,
    ChannelRequiredError,
    UrlRequiredError,
    build_api,
    build_api_auto_spec,
)


def main(
    url: Optional[str] = typer.Option(None, envvar='ASYNCAPI_URL'),
    channel: Optional[str] = typer.Option(None, envvar='ASYNCAPI_CHANNEL'),
    server: str = typer.Option('development', envvar='ASYNCAPI_SERVER'),
    operations_module: str = typer.Option(
        '', envvar='ASYNCAPI_OPERATIONS_MODULE'
    ),
    operation_id: Optional[str] = typer.Option(
        None, envvar='ASYNCAPI_OPERATION_ID'
    ),
    server_url: Optional[str] = typer.Option(
        None, envvar='ASYNCAPI_SERVER_URL'
    ),
    republish_errors: bool = typer.Option(
        True, envvar='ASYNCAPI_REPUBLISH_ERRORS'
    ),
) -> None:
    fork_server()

    if channel is None:
        raise ChannelRequiredError()

    if server_url and operation_id and operations_module:
        api = build_api_auto_spec(
            server_url,
            channel,
            operations_module,
            operation_id,
            republish_errors=republish_errors,
        )

    elif url is None:
        raise UrlRequiredError()

    else:
        api = build_api(
            url,
            server=server,
            operations_module=operations_module,
            republish_errors=republish_errors,
        )

    graceful_stop()
    loop = get_event_loop()
    start(loop, api, channel)
    loop.run_forever()


def start(loop: AbstractEventLoop, api: AsyncApi, channel: str) -> None:
    async def init() -> None:
        await api.broadcast.connect()
        await api.listen(channel)

    task = loop.create_task(init())
    task.add_done_callback(task_callback)

    typer.echo('Waiting messages...')


def graceful_stop() -> None:
    def exit_gracefully(signum: int = -1, frame: Any = None) -> None:
        signame = {  # type: ignore
            signal.SIGINT: 'SIGINT',
            signal.SIGTERM: 'SIGTERM',
        }.get(signum)
        typer.echo(f'Received {signame} scheduling shutdown...')

        time.sleep(1)
        raise SystemExit(0)

        signal.signal(signal.SIGINT, exit_gracefully)
        signal.signal(signal.SIGTERM, exit_gracefully)


def task_callback(future: Any) -> None:
    future.result()  # pragma: no cover


def run() -> None:
    typer.run(main)  # pragma: no cover


def fork_server() -> None:
    workers = int(os.environ.get('ASYNCAPI_WORKERS', 1))

    if workers > 1:
        typer.echo(f'Forking for {workers} workers')
        for _ in range(workers - 1):
            pid = os.fork()
            if pid == 0:
                break



class AsyncApiError(Exception):
    ...


class ReferenceNotFoundError(AsyncApiError):
    ...


class InvalidChannelError(AsyncApiError):
    ...


class OperationIdNotFoundError(AsyncApiError):
    ...


class ChannelOperationNotFoundError(AsyncApiError):
    ...


class UrlRequiredError(AsyncApiError):
    ...


class ChannelRequiredError(AsyncApiError):
    ...



import dataclasses
from enum import Enum
from typing import Any, Dict, Optional


@dataclasses.dataclass
class Info:
    title: str
    version: str
    description: str


class ProtocolType(Enum):
    KAFKA = 'kafka'
    REDIS = 'redis'
    POSTGRES = 'postgres'


@dataclasses.dataclass
class Server:
    name: str
    url: str
    protocol: ProtocolType
    description: str


@dataclasses.dataclass
class Message:
    name: str
    title: str
    summary: str
    content_type: str
    payload: Any


@dataclasses.dataclass
class Subscribe:
    message: Message
    operation_id: Optional[str] = None


@dataclasses.dataclass
class Channel:
    name: str
    description: str
    subscribe: Subscribe


@dataclasses.dataclass
class Components:
    messages: Optional[Dict[str, Message]] = None
    schemas: Optional[Dict[str, Any]] = None


@dataclasses.dataclass
class Specification:
    info: Info
    servers: Dict[str, Server]
    channels: Dict[str, Channel]
    components: Optional[Components] = None
    default_content_type: Optional[str] = None



import asyncio
import dataclasses
from typing import Any, Callable, Dict, Tuple

from broadcaster import Broadcast
from jsonschema import validate

from .entities import Specification
from .exceptions import (
    ChannelOperationNotFoundError,
    InvalidChannelError,
    OperationIdNotFoundError,
)


OperationsTypeHint = Dict[Tuple[str, str], Callable[..., Any]]


@dataclasses.dataclass
class AsyncApi:
    spec: Specification
    operations: OperationsTypeHint
    broadcast: Broadcast
    republish_error_messages: bool = True

    async def publish(self, channel_id: str, message: Dict[str, Any]) -> None:
        self.validate_message(channel_id, message)
        await self.broadcast.publish(channel=channel_id, message=message)

    async def listen(self, channel_id: str) -> None:
        try:
            operation_id = self.spec.channels[
                channel_id
            ].subscribe.operation_id
        except KeyError:
            raise InvalidChannelError(channel_id)

        if operation_id is None:
            raise ChannelOperationNotFoundError(channel_id)

        async with self.broadcast.subscribe(channel=channel_id) as subscriber:
            async for event in subscriber:
                self.validate_message(channel_id, event.message)

                try:
                    coro = self.operations[(channel_id, operation_id)](
                        event.message
                    )

                    if asyncio.iscoroutine(coro):
                        await coro

                except KeyError:
                    raise OperationIdNotFoundError(operation_id)

                except Exception:
                    if not self.republish_error_messages:
                        raise

                    await self.publish(channel_id, event.message)

    def validate_message(
        self, channel_id: str, message: Dict[str, Any]
    ) -> None:
        try:
            schema = self.spec.channels[channel_id].subscribe.message.payload
        except KeyError:
            raise InvalidChannelError(channel_id)

        validate(instance=message, schema=schema)
